import re
import inspect
from functools import wraps, partial
from pydantic import ValidationError, BaseModel
import falcon

from falibrary.config import Config
from falibrary.route import OpenAPI, RedocPage
from falibrary.utils import find_routes, parse_path, get_summary_desc


class Falibrary:
    """
    :param app: Falcon instance [optional](you can register it later)
    :param kwargs: key-value for config, see :class:`falibrary.config.Config`
    """

    def __init__(self, app=None, **kwargs):
        self.app = app
        self.models = {}
        self.config = Config()
        for key, value in kwargs.items():
            setattr(self.config, key.upper(), value)

        self.STATUS = re.compile(r'(?P<code>^\d{3}) (?P<msg>[\w ]+$)')
        if self.app:
            assert isinstance(app, falcon.API)
            self._register_route()

    def register(self, app):
        """
        :param app: Falcon instance

        register this library to Falcon application to get routes
        """
        assert isinstance(app, falcon.API)
        self.app = app
        self._register_route()

    def update_config(self, **kwargs):
        """
        update config

        this can be done before generate the APIspecs
        """
        for key, value in kwargs.items():
            setattr(self.config, key.upper(), value)

    def validate(self, query=None, data=None, resp=None, x=[]):
        """
        validate query, JSON data, and response according to
        ``pydantic.BaseModel``

        :param query: Schema for query args
        :param data: Schema for JSON data
        :param response: Schema for JSON response
        :param x: List of :class:`falcon.status_codes`

        .. code-block:: python

            from falibrary imoprt Falibrary
            from wsgiref import simple_server
            from pydantic import BaseModel, Schema

            class Query(BaseModel):
                text: str
                limit: int

            api = Falibrary(title='demo', version='0.1')

            class Demo:
                @api.vaildate(query=Query, x=[falcon.HTTP_422])
                def on_post(self, req, resp):
                    print(req.content.query)
                    raise falcon.HTTPUnprocessableEntity()

        """
        def decorator_validation(func):
            @wraps(func)
            def validation(self, _req, _resp, *args, **kwargs):
                try:
                    if query:
                        setattr(_req.context, 'query', query(**_req.params))
                    if data:
                        setattr(_req.context, 'data', data(**_req.media))
                except ValidationError as err:
                    raise falcon.HTTPUnprocessableEntity(
                        'Schema failed validation',
                        description=str(err),
                    )
                except Exception:
                    raise

                response = func(self, _req, _resp, *args, **kwargs)
                if resp:
                    _resp.media = response.dict()

                return response

            # register ``pydantic.BaseModel``
            for name, model in zip(
                ('query', 'data', 'resp'), (query, data, resp)
            ):
                if model:
                    assert issubclass(model, BaseModel)
                    self.models[model.__name__] = model.schema()
                    setattr(validation, name, model.__name__)

            # handle exceptions
            code_msg = {}
            for exception in x:
                match = self.STATUS.match(exception)
                assert match
                code_msg[match.group('code')] = match.group('msg')

            if code_msg:
                validation.x = code_msg

            # register decorator
            validation._decorator = self

            return validation
        return decorator_validation

    def _register_route(self):
        """
        register doc page and OpenAPI spec file
        """
        self.config.SPEC_URL = f'/{self.config.PATH}/{self.config.FILENAME}'
        self.app.add_route(
            f'/{self.config.PATH}',
            RedocPage(self.config)
        )
        self.app.add_route(
            self.config.SPEC_URL,
            OpenAPI(self)
        )

    @property
    def spec(self):
        """
        get the spec of API document
        """
        if not hasattr(self, '_spec'):
            self._generate_spec()
        return self._spec

    def bypass(self, func):
        if self.config.MODE == 'greedy':
            return False
        elif self.config.MODE == 'strict':
            if getattr(func, '_decorator', None) == self:
                return False
            return True
        else:
            decorator = getattr(func, '_decorator', None)
            if decorator and decorator != self:
                return True
            return False

    def _generate_spec(self):
        routes = {}
        tags = {}
        assert self.config.MODE in self.config._SUPPORT_MODE
        for route in find_routes(self.app._router._roots):
            path, parameters = parse_path(route.uri_template)
            routes[path] = {}
            for method, func in route.method_map.items():
                if isinstance(func, partial):
                    # ignore exception handlers
                    continue

                if self.bypass(func):
                    continue

                name = route.resource.__class__.__name__
                if name not in tags:
                    tags[name] = {
                        'name': name.lower(),
                        'description': inspect.getdoc(route.resource) or '',
                    }
                summary, desc = get_summary_desc(func)
                spec = {
                    'summary': summary or f'{name} <{method}>',
                    'operationID': f'{name}__{method.lower()}',
                    'description': desc or '',
                    'tags': [name.lower()],
                }

                if hasattr(func, 'data'):
                    spec['requestBody'] = {
                        'content': {
                            'application/json': {
                                'schema': {
                                    '$ref': f'#/components/schemas/{func.data}'
                                }
                            }
                        }
                    }

                params = parameters[:]
                if hasattr(func, 'query'):
                    params.append({
                        'name': func.query,
                        'in': 'query',
                        'required': True,
                        'schema': {
                            '$ref': f'#/components/schemas/{func.query}',
                        }
                    })
                spec['parameters'] = params

                responses = {}
                has_2xx = False
                if hasattr(func, 'x'):
                    for code, msg in func.x.items():
                        if code.startswith('2'):
                            has_2xx = True
                        responses[code] = {
                            'description': msg,
                        }

                if hasattr(func, 'resp'):
                    responses['200'] = {
                        'description': 'Successful Response',
                        'content': {
                            'application/json': {
                                'schema': {
                                    '$ref': f'#/components/schemas/{func.resp}'
                                }
                            }
                        },
                    }
                elif not has_2xx:
                    responses['200'] = {'description': 'Successful Response'}

                if any([hasattr(func, schema)
                        for schema in ('query', 'data', 'resp')]):
                    responses['422'] = {
                        'description': 'Validation Error',
                    }

                spec['responses'] = responses

                routes[path][method.lower()] = spec

        definitions = {}
        for _, schema in self.models.items():
            if 'definitions' in schema:
                for key, value in schema['definitions'].items():
                    definitions[key] = value
                del schema['definitions']

        data = {
            'openapi': self.config.OPENAPI_VERSION,
            'info': {
                'title': self.config.TITLE,
                'version': self.config.VERSION,
            },
            'tags': list(tags.values()),
            'paths': {
                **routes
            },
            'components': {
                'schemas': {
                    name: schema for name, schema in self.models.items()
                },
            },
            'definitions': definitions
        }
        self._spec = data



"""
utils functions
"""
import re
from inspect import getdoc
from falcon.routing.compiled import _FIELD_PATTERN

from falibrary.route import _doc_class_name

# NOTE from `falcon.routing.compiled.CompiledRouterNode`
ESCAPE = r'[\.\(\)\[\]\?\$\*\+\^\|]'
ESCAPE_TO = r'\\\g<0>'
EXTRACT = r'{\2}'
# NOTE this regex is copied from werkzeug.routing._converter_args_re and
# modified to support only int args
INT_ARGS = re.compile(r'''
    ((?P<name>\w+)\s*=\s*)?
    (?P<value>\d+)\s*
''', re.VERBOSE)
INT_ARGS_NAMES = ('num_digits', 'min', 'max')


def find_routes(root):
    routes = []

    def find_node(node):
        if node.resource and node.resource.__class__.__name__ not in _doc_class_name:
            routes.append(node)

        for child in node.children:
            find_node(child)

    for route in root:
        find_node(route)

    return routes


def get_summary_desc(func):
    doc = getdoc(func)
    if doc is None:
        return None, None
    doc = doc.split('\n\n', 1)
    if len(doc) == 1:
        return doc[0], None
    return doc


def parse_path(path):
    subs, parameters = [], []
    for segment in path.strip('/').split('/'):
        matches = _FIELD_PATTERN.finditer(segment)
        if not matches:
            subs.append(segment)
            continue

        escaped = re.sub(ESCAPE, ESCAPE_TO, segment)
        subs.append(_FIELD_PATTERN.sub(EXTRACT, escaped))

        for field in matches:
            variable, converter, argstr = [field.group(name) for name in
                                           ('fname', 'cname', 'argstr')]

            if converter == 'int':
                if argstr is None:
                    argstr = ''

                arg_values = [None, None, None]
                for index, field in enumerate(INT_ARGS.finditer(argstr)):
                    name, value = field.group('name'), field.group('value')
                    if name:
                        index = INT_ARGS_NAMES.index(name)
                    arg_values[index] = value

                num_digits, minumum, maximum = arg_values
                schema = {
                    'type': 'integer',
                    'format': f'int{num_digits}' if num_digits else 'int32',
                }
                if minumum:
                    schema['minimum'] = minumum
                if maximum:
                    schema['maximum'] = maximum
            elif converter == 'uuid':
                schema = {
                    'type': 'string',
                    'format': 'uuid'
                }
            elif converter == 'dt':
                schema = {
                    'type': 'string',
                    'format': 'date-time',
                }
            else:
                # no converter specified or customized converters
                schema = {'type': 'string'}

            parameters.append({
                'name': variable,
                'in': 'path',
                'required': True,
                'schema': schema,
            })

    return f'/{"/".join(subs)}', parameters



class Config:
    """
    :ivar MODE: mode for route. **normal** includes undecorated routes and
        routes decorated by this instance. **strict** only includes routes
        decorated by this instance. **greedy** includes all the routes.
    :ivar PATH: path for API document page
    :ivar UI: UI for API document, 'redoc' or 'swagger'
    :ivar OPENAPI_VERSION: OpenAPI version
    :ivar TITLE: service name
    :ivar VERSION: service version
    :ivar DOMAIN: service host domain
    """

    def __init__(self):
        self.PATH = 'apidoc'
        self.TEMPLATE_FOLDER = 'templates'
        self.FILENAME = 'openapi.json'
        self.OPENAPI_VERSION = '3.0.2'
        self.UI = 'redoc'
        self._SUPPORT_UI = {'redoc', 'swagger'}
        self.MODE = 'normal'
        self._SUPPORT_MODE = {'normal', 'strict', 'greedy'}

        self.TITLE = 'Service API Document'
        self.VERSION = '0.1'
        self.DOMAIN = None
